from async_sqs_consumer.queue import (
    delete_messages,
    get_queue_messages,
)
from async_sqs_consumer.resources import (
    sqs_shutdown,
    sqs_startup,
)
from async_sqs_consumer.utils import (
    validate_message,
)
from async_sqs_consumer.utils.retry import (
    retry_call,
)
import asyncio
from contextlib import (
    suppress,
)
from functools import (
    partial,
)
import json
import logging
import signal
import sys
from typing import (
    Any,
    Callable,
    Literal,
    Optional,
)

logging.basicConfig(level=logging.INFO)

LOGGER = logging.getLogger(__name__)


class Worker:
    def __init__(self, queue_url: str) -> None:
        self.queue_url = queue_url
        self.running = False
        self.handlers: dict[str, Callable[..., Any]] = {}
        self._events: dict[str, list[Callable[[], None]]]
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._main_task: Optional[asyncio.Task] = None

    def _sigint_handler(self, *_args: Any) -> None:
        sys.stdout.write("\b\b\r")
        sys.stdout.flush()
        logging.getLogger("system").warning(
            "Received <ctrl+c> interrupt [SIGINT]"
        )
        self._stop_worker()

    def _sigterm_handler(self, *_args: Any) -> None:
        logging.getLogger("system").warning(
            "Received termination signal [SIGTERM]"
        )
        self._stop_worker()

    def _stop_worker(self, *_: Any) -> None:
        if self._loop:
            self._loop.create_task(sqs_shutdown())
            LOGGER.debug("Stop worker")
            self.running = False
            for event in self._events.get("shutdown", []):
                self._loop.run_until_complete(event())
            if self._main_task:
                self._main_task.cancel()

    def task(self, task_name: str) -> Callable[[Callable[[], None]], None]:
        def decorator(func: Callable[[], Any]) -> None:
            self.handlers[task_name] = func

        return decorator

    def on_event(
        self, event_name: Literal[""]
    ) -> Callable[[Callable[[], None]], None]:
        def decorator(func: Callable[[], None]) -> None:
            if event_name not in self._events:
                self._events[event_name] = [func]
            else:
                self._events[event_name] = [*self._events[event_name], func]

        return decorator

    def start(self) -> None:
        self.running = True
        loop = asyncio.get_event_loop()
        self._loop = loop

        if self._loop and self._loop.is_closed():
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        for signame in ("SIGINT", "SIGTERM"):
            self._loop.add_signal_handler(
                getattr(signal, signame), partial(self._stop_worker, self)
            )

        signal.siginterrupt(signal.SIGTERM, False)
        signal.siginterrupt(signal.SIGUSR1, False)
        signal.signal(signal.SIGINT, partial(self._sigint_handler, self))
        signal.signal(signal.SIGTERM, partial(self._sigterm_handler, self))

        self._loop.run_until_complete(sqs_startup())
        self._main_task = loop.create_task(self.consumer())
        if self._main_task:
            self._main_task.add_done_callback(partial(self._stop_worker, self))

        for event in self._events.get("startup", []):
            self._loop.run_until_complete(event())

        self._loop.run_until_complete(self._main_task)

    def _delete_message(
        self, _future: asyncio.Future, *, receipt_handle: Any
    ) -> None:
        asyncio.ensure_future(
            delete_messages(self.queue_url, receipt_handle),
            loop=self._loop,
        )

    async def _process_message(self, message_content: dict[str, Any]) -> None:
        try:
            body = json.loads(message_content["Body"])
        except json.JSONDecodeError:
            return

        if not validate_message(body):
            LOGGER.error("Failed to validate message")
            return

        handler = self.handlers.get(body["task"])
        if not handler:
            LOGGER.warning(
                "The task %s has not a candidate to be proceeded",
                body["task_name"],
            )
            return
        if not self._loop:
            LOGGER.warning("The main event loop is not initialized")
            return
        task = self._loop.create_task(
            retry_call(
                handler,
                fargs=body.get("args", []),
                fkwargs=body.get("kwargs", {}),
                tries=body.get("retries", 1),
            )
        )
        task.add_done_callback(
            partial(
                self._delete_message,
                receipt_handle=message_content["ReceiptHandle"],
            )
        )

    async def consumer(self) -> None:
        while self.running:
            with suppress(asyncio.CancelledError):
                messages = await get_queue_messages(self.queue_url)
                for message in messages:
                    await self._process_message(message)



MESSAGE_SCHEMA = {
    "type": "object",
    "properties": {
        "task": {"type": "string"},
        "id": {"type": "string"},
        "args": {"type": "array"},
        "kwargs": {"type": "object"},
        "retries": {"type": "number"},
        "eta": {"type": "string"},
        "expires": {"type": "string"},
    },
    "required": ["task", "id"],
}



from aiohttp.client_exceptions import (
    ClientConnectorError,
    ServerDisconnectedError,
)
from async_sqs_consumer.resources import (
    get_sqs_client,
)
from async_sqs_consumer.utils.retry import (
    retry,
)
import asyncio
from botocore.exceptions import (
    ClientError,
)
from typing import (
    Any,
)

NETWORK_ERRORS = (
    ServerDisconnectedError,
    ClientConnectorError,
    ClientError,
    asyncio.TimeoutError,
)


@retry(exceptions=NETWORK_ERRORS, tries=3, delay=0.2)
async def get_queue_messages(queue_url: str) -> list[dict[str, Any]]:
    client = await get_sqs_client()
    response = await client.receive_message(
        QueueUrl=queue_url,
        MaxNumberOfMessages=10,
        VisibilityTimeout=600,
    )
    return response.get("Messages", [])


@retry(exceptions=NETWORK_ERRORS, tries=3, delay=0.2)
async def delete_messages(
    queue_url: str, receipt_handle: dict[str, Any]
) -> None:
    client = await get_sqs_client()
    await client.delete_message(
        ReceiptHandle=receipt_handle,
        QueueUrl=queue_url,
    )



import functools
from functools import (
    partial,
)
import logging
import random
from time import (
    sleep,
)
import traceback

logging_logger = logging.getLogger(__name__)


async def __retry_internal(  # pylint: disable=too-many-arguments
    func,
    exceptions=Exception,
    tries=-1,
    delay=0,
    max_delay=None,
    backoff=1,
    jitter=0,
    logger=logging_logger,
    log_traceback=False,
    on_exception=None,
):
    _tries, _delay = tries, delay
    while _tries:
        try:
            return await func()
        except exceptions as exc:
            if on_exception is not None:
                if on_exception(exc):
                    break

            _tries -= 1
            if not _tries:
                raise

            if logger is not None:
                try:
                    func_qualname = func.func.__qualname__
                except AttributeError:
                    func_qualname = str(func.func)
                logger.warning(
                    "%s: %s in %s.%s, retrying in %s seconds...",
                    exc.__class__.__qualname__,
                    exc,
                    func.func.__module__,
                    func_qualname,
                    _delay,
                )
                if log_traceback:
                    logger.warning(traceback.format_exc())

            sleep(_delay)
            _delay *= backoff

            if isinstance(jitter, tuple):
                _delay += random.uniform(*jitter)
            else:
                _delay += jitter

            if max_delay is not None:
                _delay = min(_delay, max_delay)


def retry(
    exceptions=Exception,
    tries=-1,
    delay=0,
    max_delay=None,
    backoff=1,
    jitter=0,
    logger=logging_logger,
    log_traceback=False,
    on_exception=None,
):
    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*fargs, **fkwargs):
            args = fargs if fargs else []
            kwargs = fkwargs if fkwargs else {}
            return await __retry_internal(
                partial(func, *args, **kwargs),
                exceptions,
                tries,
                delay,
                max_delay,
                backoff,
                jitter,
                logger,
                log_traceback,
                on_exception,
            )

        return wrapper

    return decorator


async def retry_call(
    func,
    fargs=None,
    fkwargs=None,
    exceptions=Exception,
    tries=-1,
    delay=0,
    max_delay=None,
    backoff=1,
    jitter=0,
    logger=logging_logger,
):
    args = fargs if fargs else []
    kwargs = fkwargs if fkwargs else {}
    return await __retry_internal(
        partial(func, *args, **kwargs),
        exceptions,
        tries,
        delay,
        max_delay,
        backoff,
        jitter,
        logger,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['async_sqs_consumer', 'async_sqs_consumer.utils']

package_data = \
{'': ['*']}

install_requires = \
['aioboto3>=10.1.0,<11.0.0',
 'aiobotocore>=2.4.0,<3.0.0',
 'click>=8.1.3,<9.0.0',
 'jsonschema>=4.16.0,<5.0.0']

setup_kwargs = {
    'name': 'async-sqs-consumer',
    'version': '0.1.1',
    'description': '',
    'long_description': '# Async SQS consumer\n\nThis is a simple asynchronous worker for consuming messages from AWS SQS\n\n### Usage\n```python\nfrom async_sqs_consumer.worker import (\n    Worker,\n)\n\nworker = Worker(\n    queue_url="https://sqs.us-east-1.amazonaws.com/xxxxxxx/queue_name"\n)\n\n\n@worker.task("report")\nasync def report(text: str) -> None:\n    print(text)\n\nworker.start()\n\n```\n',
    'author': 'Diego Restrepo',
    'author_email': 'drestrepo@fluidattacks.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/drestrepom/async_sqs_consumer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
